/*
 * Application startup and shutdown.
 */

#include <PalmOS.h>

#include "flipMain.h"
#include "flipDB.h"
#include "flipFlip.h"
#include "flipList.h"
#include "flipEdit.h"
#include "flipDicts.h"
#include "flipAbout.h"
#include "resdefs.h"


/*
 * Global variables
 */

/* The currently open word database. */
DmOpenRef wordDB;


/*
 * Application stuff
 */

static struct FormInfo *GetFormInfo (UInt16 formId)
{
	switch (formId) {
	
	case FlipForm: 
		return &FlipFormInfo;
	case WordListForm: 
		return &ListFormInfo;
	case WordEditForm: 
		return &EditFormInfo;
	case DictForm: 
		return &DictFormInfo;
	case AboutForm: 
		return &AboutFormInfo;
	
	default: return NULL;
	}
}

static void LoadPrefs (FlipAppPrefs *prefs)
{
	Int16 found;
	UInt16 size;

	found = PrefGetAppPreferences(flipCreatorId, FLIP_APP_PREFS_ID, 
				      prefs, &size, true);
	if (found == noPreferenceFound || size < sizeof(FlipAppPrefs)) {
		prefs->dbIndex = 0;
		prefs->flipWord = 0;
		prefs->flipFlipped = false;
	}
}

static void SavePrefs (FlipAppPrefs *prefs)
{
	PrefSetAppPreferences(flipCreatorId, FLIP_APP_PREFS_ID, FLIP_VERSION, 
			      prefs, sizeof(FlipAppPrefs), true);
}

static Err StartApplication(void)
{
	Err err;
	FlipAppPrefs prefs;
	FlipDBInfo dbInfo;

	LoadPrefs(&prefs);

	err = FlipGetDatabase(prefs.dbIndex, &dbInfo);
	if (err)
		return err;

//	err = OpenOrCreateDB(&wordDB, dmModeReadWrite);
	err = SwitchDB(&wordDB, &dbInfo);
	if (err)
		return err;

	FlipWord(prefs.flipWord, prefs.flipFlipped);

	return errNone;
}

static void StopApplication(void)
{
	Err err;
	FlipAppPrefs prefs;
	FlipDBInfo dbInfo;

	err = FlipGetOpenDatabase(wordDB, &dbInfo);
	err = FlipGetDatabaseIndex(&prefs.dbIndex, &dbInfo);

	FlipGetFlipState(&prefs.flipWord, &prefs.flipFlipped);

	SavePrefs(&prefs);

	FrmCloseAllForms();

	DmCloseDatabase(wordDB);
}

static Boolean FormHandleEvent(EventPtr event)
{
	FormPtr frmP;
	struct FormInfo *fi;

	frmP = FrmGetActiveForm();
	fi = GetFormInfo(FrmGetActiveFormID());
	if (event->eType == frmOpenEvent) {
		if (fi && fi->InitForm)
			fi->InitForm();
		if (fi && fi->UpdateForm)
			fi->UpdateForm();
		FrmDrawForm(frmP);
		return true;
	} else if (event->eType == frmUpdateEvent) {
		if (fi && fi->UpdateForm)
			fi->UpdateForm();
		FrmDrawForm(frmP);
		return true;
	} else if (event->eType == frmCloseEvent) {
		if (fi && fi->DeinitForm)
			fi->DeinitForm();
		return false;
	} else if (fi && fi->EventHandler) {
		return fi->EventHandler(event);
	}

	return false;
}



static Boolean ApplicationHandleEvent(EventPtr event)
{
	UInt16 formId;
	FormPtr frmP;

	if (event->eType == frmLoadEvent) {
		formId = event->data.frmLoad.formID;
		frmP = FrmInitForm(formId);
		FrmSetActiveForm(frmP);
		FrmSetEventHandler(frmP, FormHandleEvent);

		return true;
	}

	return FormHandleEvent(event);
}

static void EventLoop(void)
{
	Err err;
	EventType event;
	struct FormInfo *fi;

	do {
		EvtGetEvent(&event, evtWaitForever);

		/* Intercept hard buttons etc. */
		fi = GetFormInfo(FrmGetActiveFormID());
		if (fi && fi->EventStealer && fi->EventStealer(&event))
			continue;

		if (SysHandleEvent(&event))
			continue;

		if (MenuHandleEvent(0, &event, &err))
			continue;

		if (ApplicationHandleEvent(&event))
			continue;

		FrmDispatchEvent(&event);

	} while (event.eType != appStopEvent);
}


UInt32 PilotMain(UInt16 cmd, void *cmdPBP, UInt16 launchFlags)
{
	UInt16 err;

	if (cmd == sysAppLaunchCmdNormalLaunch) {
		err = StartApplication();
		if (err)
			return err;

		EventLoop();

		StopApplication();
	}
	return 0;
}
